"""Menu TUI renderable."""
# dem/cli/tui/renderable/menu.py

from rich import live, table, align, panel, box
from readchar import readkey, key

class BaseMenu(table.Table):
    """ Base class for the menus.
    
        Shouldn't be instantiated directly.
        
        Class attributes:
            cursor_on -- text that represents the cursor
            cursor_off -- the cursor is not in this cell
    """
    cursor_on = "* "
    cursor_off = "  "

    def __init__(self, box: box.Box | None = box.HEAVY_HEAD) -> None:
        """ Construct the BaseMenu based on the rich's Table class.
        
            Args:
                box -- how to draw the table's edges (None: no outline)
        """
        super().__init__(box=box)
        self.cursor_pos = 0

    def remove_cursor(self, column: int = 0, row: int = 0) -> None:
        """ Remove the cursor from the given cell.
        
            Args:
                column -- the column index
                row -- the row index
        """
        cell = self.columns[column]._cells[row].replace(self.cursor_on, self.cursor_off, 1)
        self.columns[column]._cells[row] = cell

    def add_cursor(self, column: int = 0, row: int = 0) -> None:
        """ Add the cursor to the given cell.
        
            Args:
                column -- the column index
                row -- the row index
        """
        cell = self.columns[column]._cells[row].replace(self.cursor_off, self.cursor_on, 1)
        self.columns[column]._cells[row] = cell

    def set_title(self, title: str) -> None:
        """ Set the text above the menu.
        
            Args:
                title -- the text to set
        """
        self.title = title + "\n"

class VerticalMenu(BaseMenu):
    """ Menu with vertical navigation.
    
        Shouldn't be instantiated directly.
        
        Class attributes:
            CURSOR_UP -- move cursor upwards
            CURSOR_DOWN -- move cursor downwards
    """
    (
        CURSOR_UP,
        CURSOR_DOWN
    ) = range(2)

    def __init__(self) -> None:
        """ Construct a VerticalMenu based on the BaseMenu."""
        super().__init__()

    def move_cursor(self, cursor_direction: int) -> None:
        """ Move the cursor in the given direction.
        
            Args:
                cursor_direction -- which direction to move
        """
        self.remove_cursor()

        if (cursor_direction == self.CURSOR_UP):
            if self.cursor_pos == 0:
                self.cursor_pos = self.row_count - 1
            else:
                self.cursor_pos -= 1
        else:
            if self.cursor_pos == self.row_count - 1:
                self.cursor_pos = 0
            else:
                self.cursor_pos += 1

        self.add_cursor()
 
    def remove_cursor(self) -> None:
        """ Remove the cursor."""
        super().remove_cursor(row=self.cursor_pos)

    def add_cursor(self) -> None:
        """ Add the cursor."""
        super().add_cursor(row=self.cursor_pos)

    def handle_user_input(self, input: str) -> None:
        """ Handle user input.
        
            Args:
                input -- the user input (handle up/down arrows, j and k)
        """
        match input:
            case key.UP | 'k':
                self.move_cursor(self.CURSOR_UP)
            case key.DOWN | 'j':
                self.move_cursor(self.CURSOR_DOWN)

class HorizontalMenu(BaseMenu):
    """ Menu with horizontal navigation.
    
        Shouldn't be instantiated directly.
        
        Class attributes:
            CURSOR_LEFT -- move cursor left
            CURSOR_RIGHT -- move cursor right
    """
    (
        CURSOR_LEFT,
        CURSOR_RIGHT
    ) = range(2)

    def __init__(self) -> None:
        """ Construct a HorizontalMenu based on the BaseMenu."""
        super().__init__(box=None)

    def move_cursor(self, cursor_direction: int) -> None:
        """ Move the cursor in the given direction.
        
            Args:
                cursor_direction -- which direction to move
        """
        self.remove_cursor()

        if (cursor_direction == self.CURSOR_LEFT):
            if self.cursor_pos == 0:
                self.cursor_pos = len(self.columns) - 1
            else:
                self.cursor_pos -= 1
        else:
            if self.cursor_pos == len(self.columns) - 1:
                self.cursor_pos = 0
            else:
                self.cursor_pos += 1

        self.add_cursor()

    def remove_cursor(self) -> None:
        """ Remove the cursor."""
        super().remove_cursor(column=self.cursor_pos)

    def add_cursor(self) -> None:
        """ Add the cursor."""
        super().add_cursor(column=self.cursor_pos)

    def handle_user_input(self, input:str) -> None:
        """ Handle user input.
        
            Args:
                input -- the user input (handle left/right arrows, h and l)
        """
        match input:
            case key.LEFT | 'h':
                self.move_cursor(self.CURSOR_LEFT)
            case key.RIGHT | 'l':
                self.move_cursor(self.CURSOR_RIGHT)

class CancelNextMenu(HorizontalMenu):
    """ Horizontal menu with two items: cancel and next.
    
        Class attributes:
            menu_items -- the menu items with rich text formatting
    """
    menu_items = ("[underline]cancel[/]", "[underline]next[/]")
    
    def __init__(self) -> None:
        """ Construct a HorizontalMenu with 2 columns and 1 row."""
        super().__init__()
        self.add_row(self.cursor_off + self.menu_items[0], 
                     self.cursor_off + self.menu_items[1])
        self.is_selected = False

    def handle_user_input(self, input: str) -> None:
        """ Handle user input or pass to parent.
        
            Args:
                input -- the user input (handle enter)
        """
        if input == key.ENTER:
            self.is_selected = True
        else:
            super().handle_user_input(input)

    def get_selection(self) -> str:
        """ Get the selected item.
        
            Return with the selected item.
        """
        return self.columns[self.cursor_pos]._cells[0]

class BackMenu(HorizontalMenu):
    """ Horizontal menu with a single item: back.
    
        Class attributes:
            menu_item -- the menu item with rich text formatting
    """
    menu_item = "[underline]back[/]"

    def __init__(self) -> None:
        """ Construct a HorizontalMenu with 1 column and 1 row."""
        super().__init__()
        self.add_row(self.cursor_off + self.menu_item)
        self.is_selected = False

    def handle_user_input(self, input: str) -> None:
        """ Handle user input.
        
            Args:
                input -- the user input (handle enter)
        """
        if input is key.ENTER:
            self.is_selected = True

class ToolTypeMenu(VerticalMenu):
    """ Vertical menu for the tool type selection.
    
        Class attributes:
            selected -- selected text
            not_selected -- not selected text
    """
    selected = "[green]yes[/]"
    not_selected = "no"

    def __init__(self, supported_tool_types: list[str]) -> None:
        """ Construct a VerticalMenu with 2 columns and rows that contain the supported tool types.
        
            Args:
                supported_tool_types -- the supported tool types
        """
        super().__init__()
        self.add_column("Tool types")
        self.add_column("Selected")

        for index, tool_type in enumerate(supported_tool_types):
            if (index == 0):
                # Set the cursor indicator for the first element.
                self.add_row(self.cursor_on + tool_type, self.not_selected)
            else:
                self.add_row(self.cursor_off + tool_type, self.not_selected)

    def preset_selection(self, already_selected: list[str]) -> None:
        """ Preset the given tool types.
        
            Args:
                already_selected -- list of tool types to set as selected
        """
        for row_idx, cell in enumerate(self.columns[0]._cells):
            if cell[2:] in already_selected:
                self.columns[1]._cells[row_idx] = self.selected

    def toggle_select(self) -> None:
        """ Toggle selection at the cursor's position. """
        if (self.columns[1]._cells[self.cursor_pos] is self.selected):
            self.columns[1]._cells[self.cursor_pos] = self.not_selected
        else:
            self.columns[1]._cells[self.cursor_pos] = self.selected

    def handle_user_input(self, input: str) -> None:
        """ Handle user input or pass to parent.
        
            Args:
                input -- the user input (handle enter or space)
        """
        if (input == key.ENTER) or (input == ' '):
            self.toggle_select()
        else:
            super().handle_user_input(input)

    def get_selected_tool_types(self) -> list[str]:
        """ Get selected tool types.

            Returns with a list of tool types that are in selected state.
        """
        selected_tool_types = []
        for row_index, cell in enumerate(self.columns[1]._cells):
            if cell == self.selected:
                selected_tool_types.append(self.columns[0]._cells[row_index][2:])
        return selected_tool_types

class ToolImageMenu(VerticalMenu):
    def __init__(self, tool_images: list[list[str]]) -> None:
        super().__init__()
        self.add_column("Tool images", no_wrap=True)
        self.add_column("Availability", no_wrap=True)

        for index, tool_image in enumerate(tool_images):
            if (index == 0):
                # Set the cursor indicator for the first element.
                self.add_row("* " + tool_image[0], tool_image[1])
            else:
                self.add_row("  " + tool_image[0], tool_image[1])

        self.is_selected = False

    def handle_user_input(self, input: str) -> None:
        if input == key.ENTER:
            self.is_selected = True
        else:
            super().handle_user_input(input)

    def get_selected_tool_image(self) -> str:
        return self.columns[0]._cells[self.cursor_pos][2:]

class SelectMenu(VerticalMenu):
    def __init__(self, selection: list[str]) -> None:
        super().__init__()
        self.alignment = align.Align(self, align="center", vertical="middle")
        self.show_edge = False
        self.show_lines = False

        for index, element in enumerate(selection):
            if (index == 0):
                # Set the cursor indicator for the first element.
                self.add_row("* " + element)
            else:
                self.add_row("  " + element)

        self.alignment = align.Align(self, align="center", vertical="middle")

    def wait_for_user(self):
        with live.Live(self.alignment, refresh_per_second=8, screen=True):
            while True:
                match readkey():
                    case key.UP | 'k':
                        self.move_cursor(self.CURSOR_UP)
                    case key.DOWN | 'j':
                        self.move_cursor(self.CURSOR_DOWN)
                    case key.ENTER:
                        break
            
    def get_selected(self) -> str:
        return self.columns[0]._cells[self.cursor_pos][2:]
    
    def set_title(self, title: str) -> None:
        self.width = len(title)
        super().set_title(title)

class DevEnvStatus(panel.Panel):
    def __init__(self, tool_types: list[str]) -> None:
        self.outer_table = table.Table(box=None)
        super().__init__(self.outer_table, title="Development Environment", expand=False)
        self.aligned_renderable = align.Align(self, vertical="middle")
        
        self._fill_table(tool_types)
    
    def _fill_table(self, tool_types: list[str]) -> None:
        panel_height = 3 + len(tool_types) * 4
        self.height = panel_height

        for tool_type in tool_types:
            inner_table = table.Table(box=None)
            inner_table.add_row("[bold]" + tool_type + ":[/]")
            inner_table.add_row("<not selected>")
            inner_table.add_row("")

            self.outer_table.add_row(inner_table)

    def reset_table(self, tool_types: list[str]) -> None:
        self.outer_table = table.Table(box=None)
        self._fill_table(tool_types)
        self.renderable = self.outer_table

    def set_tool_image(self, tool_selection: dict) -> None:
        for tool_type, tool_image in tool_selection.items():
            for row in self.outer_table.columns[0].cells:
                if tool_type in row.columns[0]._cells[0]:
                    row.columns[0]._cells[1] = tool_image


"""This module represents a Development Environment."""
# dem/core/dev_env.py

from dem.core.core import Core
from dem.core.exceptions import InvalidDevEnvJson
from dem.core.tool_images import ToolImages

class DevEnv(Core):
    """ A Development Environment.
        
        Class variables:
            supported_tool_types -- supported tool types
    """
    supported_tool_types = ( 
        "build system",
        "toolchain",
        "debugger",
        "deployer",
        "test framework",
        "CI/CD server",
    )

    def _check_tool_type_support(self, descriptor: dict) -> None:
        """ Check that the Dev Env doesn't contain an unsupported tool type.
        
            Private function that gets called on instantiation.
            Args:
                descriptor -- the description of the Development Environment from the dev_env.json 
                              file
        """
        for tool in descriptor["tools"]:
            if tool["type"] not in self.supported_tool_types:
                raise InvalidDevEnvJson("The following tool type is not supported: " + tool["type"])

    def __init__(self, descriptor: dict | None = None, 
                 dev_env_to_copy: "DevEnv | None" = None) -> None:
        """ Init the DevEnv class. A new instance can be create from a descriptor or from an already
            existing DevEnv instance.
        
            Args:
                descriptor -- the description of the Development Environment from the dev_env.json 
                              file
                dev_env_to_copy -- the DevEnv instance to copy
        """
        if descriptor:
            self._check_tool_type_support(descriptor)
            self.name = descriptor["name"]
            self.tools = descriptor["tools"]
        else:
            self.name = dev_env_to_copy.name
            self.tools = dev_env_to_copy.tools

    def check_image_availability(self, tool_images: ToolImages, update_tool_images: bool = False,
                                 local_only: bool = False) -> list:
        """ Checks the tool image's availability.
        
            Updates the "image_status" key for the tool dictionary.
            Returns with the statuses of the Dev Env tool images.

            Args:
                tool_images -- the images the Dev Envs can access
                update_tool_images -- update the list of available tool images
                local_only -- only local images are used
        """
        if update_tool_images == True:
            tool_images.local.update()
            if local_only is False:
                tool_images.registry.update()

        local_tool_images = tool_images.local.elements

        if local_only is True:
            registry_tool_images = []
        else:
            registry_tool_images = tool_images.registry.elements

        image_statuses = []
        for tool in self.tools:
            tool_image_name = tool["image_name"] + ':' + tool["image_version"]
            if (tool_image_name in local_tool_images) and (tool_image_name in registry_tool_images):
                image_status = ToolImages.LOCAL_AND_REGISTRY
            elif (tool_image_name in local_tool_images):
                image_status = ToolImages.LOCAL_ONLY
            elif (tool_image_name in registry_tool_images):
                image_status = ToolImages.REGISTRY_ONLY
            else:
                image_status = ToolImages.NOT_AVAILABLE
            image_statuses.append(image_status)
            tool["image_status"] = image_status

        return image_statuses


"""Collection of dem specific exceptions."""
# dem/core/exceptions.py

class InvalidDevEnvJson(Exception):
    """Raised when the dev_env.json file is invalid."""

    base_message = "Error in dev_env.json: "

    def __init__(self, message: str) -> None:
        super().__init__(self.base_message + message)

class RegistryError(Exception):
    """Raised when the communication with registry fails."""
    pass

class ContainerEngineError(Exception):
    """Raised when there is a problem with the container engine."""

    base_message = "Container engine error: "

    def __init__(self, message: str) -> None:
        super().__init__(self.base_message + message)


"""load CLI command implementation."""
# dem/cli/command/load_cmd.py

from dem.core.dev_env import DevEnv
from dem.core.platform import DevEnvLocalSetup
from dem.cli.console import stderr
import json, os

def check_is_file_exist(param: str | None) -> bool:
    if param is not None:
        return(os.path.isfile(param))     
    else:
        return False

def load_dev_env_to_dev_env_json(dev_env_local_setup: DevEnvLocalSetup,path_to_dev_env: str) -> bool:
    raw_file = open(path_to_dev_env, "r")   
            
    try:
        dev_env = json.load(raw_file)
        
        if dev_env_local_setup.get_dev_env_by_name(dev_env["name"]) is not None:
            stderr.print("[red]Error: The Development Environment exist.[/]")
            return False                       
        else:        
            new_dev_env = DevEnv(dev_env)
            dev_env_local_setup.local_dev_envs.append(new_dev_env)
            new_dev_env.check_image_availability(dev_env_local_setup.tool_images)
            dev_env_local_setup.pull_images(new_dev_env.tools)
    except json.decoder.JSONDecodeError:
       stderr.print("[red]Error: invalid json format.[/]")
       return False                       
    else:
        raw_file.close()
        return True

def execute(path_to_dev_env: str) -> None:
    platform = DevEnvLocalSetup()    

    if check_is_file_exist(path_to_dev_env) is True:                
        retval = load_dev_env_to_dev_env_json(platform,path_to_dev_env)        
        if retval == True:
            platform.flush_to_file()
        else:
            stderr.print("[red]Error: Something went wrong.[/]")    
    else:
        stderr.print("[red]Error: The input file does not exist.[/]")


"""Development Environment Catalog."""
# dem/core/dev_env_catalog.py

from dem.core.dev_env import DevEnv
from dem.core.data_management import ConfigFile
from dem.core.core import Core
import requests

class DevEnvCatalog():
    """ Development Environment Catalog. """
    def __init__(self, catalog_config: dict) -> None:
        """ Init the class with the DevEnvs available in the catalog. 

            Args:
                url -- the url of the catalog
        """
        self.config: dict = catalog_config
        self.url: str = catalog_config["url"]
        self.dev_envs: list[DevEnv] = []
        for dev_env_descriptor in requests.get(self.url, timeout=1).json()["development_environments"]:
            self.dev_envs.append(DevEnv(descriptor=dev_env_descriptor))

    def get_dev_env_by_name(self, dev_env_name: str) -> DevEnv | None:
        """ Get the Development Environment by name.
        
            Args:
                dev_env_name -- name of the Development Environment to get
            Return with the instance representing the Development Environment. If the Development 
            Environment doesn't exist in the catalog, return with None.
        """
        for dev_env in self.dev_envs:
            if dev_env.name == dev_env_name:
                return dev_env

class DevEnvCatalogs(Core):
    """ List of the available Development Environment Catalogs. """
    def __init__(self, config_file: ConfigFile) -> None:
        """ Init the class with the catalogs from the config file.

            Args:
                config_file -- contains the catalog descriptions
            """
        self._config_file: ConfigFile = config_file
        self.catalogs: list[DevEnvCatalog] = []
        for catalog_config in config_file.catalogs:
            self._try_to_add_catalog(catalog_config)

    def _try_to_add_catalog(self, catalog_config: dict) -> bool:
        try:
            self.catalogs.append(DevEnvCatalog(catalog_config))
        except Exception as e:
            self.user_output.error(str(e))
            self.user_output.error("Error: Couldn't add this Development Environment Catalog.")
            return False
        else:
            return True

    def add_catalog(self, catalog_config: dict) -> None:
        """ Add a new catalog.
        
            Args:
                catalog_config -- the new catalog to add
        """
        if self._try_to_add_catalog(catalog_config):
            self._config_file.catalogs.append(catalog_config)
            self._config_file.flush()

    def list_catalog_configs(self) -> list[dict]:
        """ List the catalog configs. (As stored in the config file.)
        
            Return with the list of the available catalog configurations.
        """
        return self._config_file.catalogs

    def delete_catalog(self, catalog_config: dict) -> None:
        """ Delete the catalog.
        
            Args:
                catalog_config -- config of the catalog to delete
        """
        for catalog in self.catalogs.copy():
            if catalog.config == catalog_config:
                self.catalogs.remove(catalog)

        self._config_file.catalogs.remove(catalog_config)
        self._config_file.flush()


"""CLI command implementation for adding a catalog."""
# dem/cli/command/add_cat_cmd.py

from dem.core.platform import DevEnvLocalSetup
from dem.cli.console import stdout

def execute(name: str, url:str) -> None:
    """ Add a new Dev Env Catalog.
    
        Args:
            name -- name of the catalog
            url -- URL of the catalog's JSON file
    """
    platform = DevEnvLocalSetup()
    catalog_config = {
        "name": name,
        "url": url
    }
    if catalog_config not in platform.dev_env_catalogs.list_catalog_configs():
        platform.dev_env_catalogs.add_catalog(catalog_config)
    else:
        stdout.print("[yellow]The input catalog is already added.[/]")


"""CLI command implementation for deleting a Dev Env Catalog."""
# dem/cli/command/del_cat_cmd.py

from dem.core.platform import DevEnvLocalSetup
from dem.cli.console import stdout, stderr

def execute(catalog_name: str) -> None:
    """ Delete the Dev Env Catalog.
        Args:
            catalog_name -- name of the catalog to delete
    """
    platform = DevEnvLocalSetup()

    for catalog_config in platform.dev_env_catalogs.list_catalog_configs():
        if catalog_config["name"] == catalog_name:
            platform.dev_env_catalogs.delete_catalog(catalog_config)
            stdout.print("[green]The input catalog has been successfully deleted.")
            break
    else:
        stderr.print("[red]Error: The input catalog name doesn't exist![/]")


"""Image selector panel."""
# dem/cli/tui/panel/image_selector.py

from dem.cli.tui.renderable.menu import ToolImageMenu, BackMenu, DevEnvStatus
from rich.panel import Panel
from rich.layout import Layout
from rich.console import RenderableType, Group
from rich.align import Align
from rich.live import Live
from readchar import readkey, key

class NavigationHint(Panel):
    hint_text = """
- [bold]move cursor[/]: arrows or vi mode
- [bold]select[/]: space or enter
- [bold]jump to back[/]: tab
"""

    def __init__(self) -> None:
        super().__init__(self.hint_text, title="Navigation", expand=False)
        self.aligned_renderable = Align(self, align="center")

class ToolImageSelectorPanel():
    def __init__(self, elements: list[list[str]], tool_types: list[str]) -> None:
        # Panel content
        self.tool_image_menu = ToolImageMenu(elements)
        self.dev_env_status = DevEnvStatus(tool_types)
        self.back_menu = BackMenu()
        self.navigation_hint = NavigationHint()

        self.menus = Align(Group(
            Align(self.tool_image_menu, align="center", vertical="middle"),
            Align(self.back_menu, align="center", vertical="middle"),
        ), align="center", vertical="middle")

        self.layout = Layout(name="root")

        self.layout.split_column(
            Layout(name="main"),
            Layout(name="navigation_hint", size=7),
        )
        self.layout["main"].split_row(
            Layout(name="menus"),
            Layout(name="dev_env_status", size=30)
        )

        self.layout["menus"].update(self.menus)
        self.layout["dev_env_status"].update(self.dev_env_status.aligned_renderable)
        self.layout["navigation_hint"].update(self.navigation_hint.aligned_renderable)

        self.back_menu.remove_cursor()
        self.active_menu = self.tool_image_menu

    def wait_for_user(self) -> None:
        with Live(self.layout, refresh_per_second=8, screen=True):
            while self.active_menu.is_selected is False:
                input = readkey()
                if input is key.TAB:
                    self.active_menu.remove_cursor()

                    if self.active_menu is self.tool_image_menu:
                        self.active_menu = self.back_menu
                    else:
                        self.active_menu = self.tool_image_menu

                    self.active_menu.add_cursor()
                else:
                    self.active_menu.handle_user_input(input)


"""Tool selector panel."""
# dem/cli/tui/panel/tool_selector.py

from dem.cli.tui.renderable.menu import ToolTypeMenu, CancelNextMenu
from rich.panel import Panel
from rich.layout import Layout
from rich.console import RenderableType, Group
from rich.live import Live
from rich.align import Align
from readchar import readkey, key

class NavigationHint():
    hint_test = """
- [bold]move cursor[/]: arrows or vi mode
- [bold]select[/]: space or enter
- [bold]jump to next/cancel[/]: tab
- [bold]finish selection[/]: press enter when [italic]next[/] is selected
"""

    def __init__(self) -> None:
        self.panel = Panel(self.hint_test, title="Navigation", expand=False)
        self.aligned_panel = Align(self.panel, align="center")

    def get_renderable(self) -> RenderableType:
        return self.aligned_panel

class ToolTypeSelectorPanel():
    def __init__(self, elements: list[str]) -> None:
        # Panel content
        self.tool_type_menu = ToolTypeMenu(elements)
        self.cancel_next_menu = CancelNextMenu()
        self.cancel_next_menu.remove_cursor()

        self.menus = Align(Group(
            Align(self.tool_type_menu, align="center", vertical="middle"),
            Align(self.cancel_next_menu, align="center", vertical="middle"),
        ), align="center", vertical="middle")

        self.navigation_hint = NavigationHint()

        self.layout = Layout(name="root")
        self.layout.split(
            Layout(name="menus"),
            Layout(name="info", size=8),
        )
        self.layout["menus"].update(self.menus)
        self.layout["info"].update(self.navigation_hint.get_renderable())

        self.no_tool_type_selected_error = Align(Panel("[yellow]You need to select at least one tool type![/]"),
                                                 align="center", vertical="middle")
        
        self.active_menu = self.tool_type_menu

    def wait_for_user(self) -> None:
        with Live(self.layout, refresh_per_second=8, screen=True):
            selection = ""
            is_error_presented = False
            while selection == "":
                input = readkey()
                if input is key.TAB:
                    if is_error_presented is True:
                        self.layout["info"].update(self.navigation_hint.get_renderable())
                    self.active_menu.remove_cursor()

                    if self.active_menu is self.tool_type_menu:
                        self.active_menu = self.cancel_next_menu
                    else:
                        self.active_menu = self.tool_type_menu

                    self.active_menu.add_cursor()
                else:
                    self.active_menu.handle_user_input(input)

                    if self.cancel_next_menu.is_selected is True:
                        selection = self.cancel_next_menu.get_selection()

                        if ("next" in selection) and (len(self.tool_type_menu.get_selected_tool_types()) == 0):
                            self.cancel_next_menu.is_selected = False
                            self.layout["info"].update(self.no_tool_type_selected_error)
                            is_error_presented = True
                            selection = ""


"""This module provides the CLI."""
# dem/cli/main.py

import typer, importlib.metadata
from typing import Optional
from dem import __command__, __app_name__
from dem.cli.command import info_cmd, list_cmd, pull_cmd, create_cmd, modify_cmd, delete_cmd, \
                            rename_cmd, clone_cmd, run_cmd, export_cmd, load_cmd, add_reg_cmd, \
                            list_reg_cmd, del_reg_cmd, add_cat_cmd, list_cat_cmd, del_cat_cmd
from dem.cli.console import stdout, stderr

typer_cli = typer.Typer(rich_markup_mode="rich")

@typer_cli.command("list")
def list_(local: bool = typer.Option(False, help="Scope is the local host."),
         all: bool = typer.Option(False, help="Scope is the organization."),
         env: bool = typer.Option(False, help="List the environments."),
         tool: bool = typer.Option(False, help="List the tool images.")) -> None:
    """
    List the Development Environments available locally or for the organization.
    
    The following option combinations are suppported:

        --local --env -> List the local Development Environments.

        --all --env -> List the organization's Development Environments.

        --local --tool -> List the local tool images.

        --all --tool -> List the tool images available in the axemsolutions registry.
    """
    list_cmd.execute(local, all, env, tool)

@typer_cli.command()
def info(dev_env_name: str = typer.Argument(...,
                                            help="Name of the Development Environment to get info about.")) -> None:
    """
    Get information about the specified Development Environment.
    """
    info_cmd.execute(dev_env_name)

@typer_cli.command()
def pull(dev_env_name: str = typer.Argument(..., 
                                            help="Name of the Development Environment to install.")) -> None:
    """
    Pull all the required tool images from the registry and install the Development Environment
    locally.
    """
    pull_cmd.execute(dev_env_name)

@typer_cli.command()
def clone(dev_env_name: str = typer.Argument(...,help="Name of the Development Environment to clone."),
           new_dev_env_name: str = typer.Argument(...,help="Name of the New Development Environment.")) -> None:
    """
    Clone existing Development Environment locally.
    """
    clone_cmd.execute(dev_env_name,new_dev_env_name)

@typer_cli.command()
def create(dev_env_name: str = typer.Argument(..., 
                                              help="Name of the Development Environment to create."),) -> None:
    """
    Create a new Development Environment.
    """
    create_cmd.execute(dev_env_name)

@typer_cli.command()
def export(dev_env_name: str = typer.Argument(...,help="Name of the Development Environment to export."),
           path_to_export: str = typer.Argument(None,help="Path where to extract the Dev Env.")) -> None:
    """
    Export the Development Environment.
    """
    export_cmd.execute(dev_env_name,path_to_export)

@typer_cli.command()
def load(path_to_dev_env: str = typer.Argument(...,help="Path to the Dev Env to import.")) -> None:
    """
    Import the Development Environment.
    """
    load_cmd.execute(path_to_dev_env)

@typer_cli.command()
def rename(dev_env_name: str = typer.Argument(...,help="Name of the Development Environment to rename."),
           new_dev_env_name: str = typer.Argument(...,help="The new name.")) -> None:
    """
    Rename the Development Environment.
    """
    rename_cmd.execute(dev_env_name,new_dev_env_name)

@typer_cli.command()
def modify(dev_env_name: str = typer.Argument(..., 
                                              help="Name of the Development Environment to modify.")) -> None:
    """
    Modify the tool types and required tool images for an existing Development Environment.
    """
    modify_cmd.execute(dev_env_name)

@typer_cli.command()
def delete(dev_env_name: str = typer.Argument(..., 
                                              help="Name of the Development Environment to delete.")) -> None:
    """
    Delete the Development Environment from the local setup. If a tool image is not required
    anymore by any of the available local Development Environments, the DEM asks the user whether
    they want to delete that image or not.
    """
    delete_cmd.execute(dev_env_name)

@typer_cli.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def run(dev_env_name: str = typer.Argument(..., 
                                           help="Run the container in this Development Environment context"),
        ctx: typer.Context = typer.Option(None)) -> None:
    """
    Run the `docker run` command in the Development Environment's context with the given parameters.  

    This command can be used as the docker CLI one, except as first argument the name of the 
    Development Environment must be set. 
    Example: dem run dev_env --name test test_image_name:latest ls -la

    See the documentation for the list of currently supported docker run parameters.
    """
    run_cmd.execute(dev_env_name, ctx.args)

@typer_cli.command()
def add_reg(name: str = typer.Argument(..., help="Name of the registry to add"), 
            url: str = typer.Argument(..., help="API URL of the registry")) -> None:
    """
    Add a new registry.

    The name of the registry is what you would normally use to pull an image. 
    Examples:
        - If the full image tag: repository/image:tag -> the name should be repository.
        - If the full image tag: 192.168.1.1:5000/image:tag -> the name should be 192.168.1.1:5000

    The URL should point to the registry's REST API. For the Docker Hub its 
    https://registry.hub.docker.com, or it can be http://localhost:5000 for a self-hosted one.
    """
    add_reg_cmd.execute(name, url)

@typer_cli.command()
def list_reg() -> None:
    """
    List the available registries.
    """
    list_reg_cmd.execute()

@typer_cli.command()
def del_reg(registry_name: str = typer.Argument(..., help="Name or IP address of the registry to delete.")) -> None:
    """
    Delete a registry.
    """
    del_reg_cmd.execute(registry_name)

@typer_cli.command()
def add_cat(name: str = typer.Argument(..., help="Name of the Development Environment Catalog to add"), 
            url: str = typer.Argument(..., help="URL of the Development Environment Catalog's JSON file")) -> None:
    """
    Add a new catalog.

    You can name the catalog as you wish. 
    
    The URL must point to an HTTP(S) server where the Catalog json file is available.
    """
    add_cat_cmd.execute(name, url)

@typer_cli.command()
def list_cat() -> None:
    """
    List the available catalogs.
    """
    list_cat_cmd.execute()

@typer_cli.command()
def del_cat(registry_name: str = typer.Argument(..., help="Name of the Development Environment Catalog to delete.")) -> None:
    """
    Delete a catalog.
    """
    del_cat_cmd.execute(registry_name)

def _version_callback(value: bool) -> None:
    if value:
        try: 
            version = importlib.metadata.version(__app_name__)
        except importlib.metadata.PackageNotFoundError:
            stdout.print("[yellow]Install DEM to get the version number.[/]")
        else: 
            stdout.print("[cyan]" + __app_name__ + " v" + version + "[/]")

        raise typer.Exit()

@typer_cli.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the dem version.",
        callback=_version_callback,
        is_eager=True,
    )) -> None:
    """
    Development Environment Manager (dem)
    
    Manage your containerized Development Environments with ease.



    ❗ Always put the input text into double quotation marks (""), if it contains whitespaces.

    """
    return


"""CLI command implementation for listing the available Development Environment Catalogs."""
# dem/cli/command/list_cat_cmd.py

from dem.core.platform import DevEnvLocalSetup
from dem.cli.console import stdout
from rich.table import Table

def execute() -> None:
    """ List available Development Environment Catalogs."""
    platform = DevEnvLocalSetup()
    catalog_config = None
    table = Table()
    table.add_column("name")
    table.add_column("url")

    for catalog_config in platform.dev_env_catalogs.list_catalog_configs():
        table.add_row(catalog_config["name"], catalog_config["url"])
    
    if catalog_config is None:
        stdout.print("[yellow]No Development Environment Catalogs are available![/]")
    else:
        stdout.print(table)


"""list CLI command implementation."""
# dem/cli/list_cmd.py

from dem.core.platform import DevEnvLocalSetup
from dem.core.dev_env import DevEnv
from dem.core.tool_images import ToolImages
from dem.cli.console import stdout, stderr
from rich.table import Table

(
    DEV_ENV_ORG_NOT_IN_REGISTRY,
    DEV_ENV_ORG_INSTALLED_LOCALLY,
    DEV_ENV_ORG_REINSTALL,
    DEV_ENV_ORG_READY,
) = range(4)

(
    DEV_ENV_LOCAL_NOT_AVAILABLE,
    DEV_ENV_LOCAL_REINSTALL,
    DEV_ENV_LOCAL_INSTALLED,
) = range(3)

dev_env_org_status_messages = {
    DEV_ENV_ORG_NOT_IN_REGISTRY: "[red]Error: Required image is not available in the registry![/]",
    DEV_ENV_ORG_INSTALLED_LOCALLY: "Installed locally.",
    DEV_ENV_ORG_REINSTALL: "Incomplete local install. The missing images are available in the registry. Use `dem pull` to reinstall.",
    DEV_ENV_ORG_READY: "Ready to be installed.",
}

dev_env_local_status_messages = {
    DEV_ENV_LOCAL_NOT_AVAILABLE: "[red]Error: Required image is not available![/]",
    DEV_ENV_LOCAL_REINSTALL: "Incomplete local install. The missing images are available in the registry. Use `dem pull` to reinstall.",
    DEV_ENV_LOCAL_INSTALLED: "Installed.",
}

def get_catalog_dev_env_status(platform: DevEnvLocalSetup, dev_env: DevEnv) -> str:
    image_statuses = dev_env.check_image_availability(platform.tool_images)
    if (ToolImages.NOT_AVAILABLE in image_statuses) or (ToolImages.LOCAL_ONLY in image_statuses):
        dev_env_status = dev_env_org_status_messages[DEV_ENV_ORG_NOT_IN_REGISTRY]
    elif (image_statuses.count(ToolImages.LOCAL_AND_REGISTRY) == len(image_statuses)) and \
            (platform.get_local_dev_env(dev_env) is not None):
        dev_env_status = dev_env_org_status_messages[DEV_ENV_ORG_INSTALLED_LOCALLY]
    else:
        if (platform.get_local_dev_env(dev_env) is not None):
            dev_env_status = dev_env_org_status_messages[DEV_ENV_ORG_REINSTALL]
        else:
            dev_env_status = dev_env_org_status_messages[DEV_ENV_ORG_READY]
    return dev_env_status

def get_local_dev_env_status(dev_env: DevEnv, tool_images: ToolImages) -> str:
    image_statuses = dev_env.check_image_availability(tool_images)
    if (ToolImages.NOT_AVAILABLE in image_statuses):
        dev_env_status = dev_env_local_status_messages[DEV_ENV_LOCAL_NOT_AVAILABLE]
    elif (ToolImages.REGISTRY_ONLY in image_statuses):
        dev_env_status = dev_env_local_status_messages[DEV_ENV_LOCAL_REINSTALL]
    else:
        dev_env_status = dev_env_local_status_messages[DEV_ENV_LOCAL_INSTALLED]
    return dev_env_status

def list_dev_envs(platform: DevEnvLocalSetup, local: bool, org: bool)-> None:
    table = Table()
    table.add_column("Development Environment")
    table.add_column("Status")

    if ((local == True) and (org == False)):
        if not platform.local_dev_envs:
            stdout.print("[yellow]No installed Development Environments.[/]")
            return
        else:
            for dev_env in platform.local_dev_envs:
                table.add_row(dev_env.name, get_local_dev_env_status(dev_env, platform.tool_images))
    elif((local == False) and (org == True)):
        if not platform.dev_env_catalogs.catalogs:
            stdout.print("[yellow]No Development Environment Catalogs are available!")
            return
        for catalog in platform.dev_env_catalogs.catalogs:
            if not catalog.dev_envs:
                stdout.print("[yellow]No Development Environments are available in the catalogs.[/]")
                return
            else:
                for dev_env in catalog.dev_envs:
                    table.add_row(dev_env.name, get_catalog_dev_env_status(platform, dev_env))
    else:
        stderr.print("[red]Error: Invalid options.[/]")
        return

    stdout.print(table)

def list_tool_images(platform: DevEnvLocalSetup, local: bool, org: bool) -> None:
    """ List tool images
    
    Args:
        local -- list local tool images
        org -- list the tool catalog
    """
    if (local == True) and (org == False):        
        local_images = platform.container_engine.get_local_tool_images()

        table = Table()
        table.add_column("Repository")
        for local_image in local_images:
            table.add_row(local_image)
        stdout.print(table)
    elif (local == False) and (org == True):
        if not platform.registries.registries:
            stdout.print("[yellow]No registries are available!")
            return

        registry_images = platform.registries.list_repos()
        if registry_images:
            table = Table()
            table.add_column("Repository")
            for registry_image in registry_images:
                table.add_row(registry_image)
            stdout.print(table)
        else:
            stdout.print("[yellow]No images are available in the registries!")

def execute(local: bool, org: bool, env: bool, tool: bool) -> None:
    platform = DevEnvLocalSetup()
    if ((local == True) or (org == True)) and (env == True) and (tool == False):
        list_dev_envs(platform, local, org)
    elif ((local == True) or (org == True)) and (env == False) and (tool == True):
        list_tool_images(platform, local, org)
    else:
        stderr.print(\
"""Usage: dem list [OPTIONS]
Try 'dem list --help' for help.

Error: You need to set the scope and what to list!""")


"""Local and registry tool images."""
# dem/core/tool_images.py

from dem.core.exceptions import RegistryError
from dem.core.container_engine import ContainerEngine
from dem.core.registry import Registries

class BaseToolImages():
    """ Base class for the tool images. 
    
        Do not instantiate it directly!
    """
    def __init__(self) -> None:
        """ Init the class. """
        self.elements = []

class LocalToolImages(BaseToolImages):
    """ Local tool images."""
    def __init__(self, container_engine: ContainerEngine) -> None:
        """ Init the class.
        
            Args: 
                container_engine -- container engine
        """
        super().__init__()
        self.container_egine = container_engine

    def update(self) -> None:
        """ Update the local tool image list using the container engine."""
        self.elements = self.container_egine.get_local_tool_images()

class RegistryToolImages(BaseToolImages):
    """ Registry tool images. """

    def __init__(self, registries: Registries) -> None:
        """ Init the class.
        
            Args: 
                registries -- registries
        """
        super().__init__()
        self._registries = registries

    def update(self) -> None:
        """ Update the list of available tools in the registry using the registry interface."""
        try:
            self.elements = self._registries.list_repos()
        except RegistryError as e:
            self.elements = []
            raise

class ToolImages():
    """ Available tool images."""
    (
        LOCAL_ONLY,
        REGISTRY_ONLY,
        LOCAL_AND_REGISTRY,
        NOT_AVAILABLE,
    ) = range(4)

    def __init__(self, container_engine: ContainerEngine, registries: Registries, 
                 update_on_instantiation: bool = True) -> None:
        """ Init the class.
        
            Args: 
                container_engine -- container engine
                registries -- registries
                update_on_instantiation -- set to false for manual update
        """
        self.local = LocalToolImages(container_engine)
        self.registry = RegistryToolImages(registries)

        if update_on_instantiation is True:
            self.local.update()
            self.registry.update()


"""Entry point for dem."""
# dem/__main__.py

from dem import __command__
from dem.cli.console import stderr, stdout
from dem.core.exceptions import RegistryError, ContainerEngineError
import dem.cli.main
import docker.errors
from dem.core.core import Core
from dem.cli.tui.tui_user_output import TUIUserOutput

def main():
    """ Entry point for the CLI application"""

    # Connect the UI to the user output interface
    Core.set_user_output(TUIUserOutput())

    try:
        dem.cli.main.typer_cli(prog_name=__command__)
    except LookupError as e:
        stderr.print("[red]" + str(e) + "[/]")
    except RegistryError as e:
        stderr.print("[red]" + str(e) + "\nUsing local tool images only![/]")
    except docker.errors.DockerException as e:
        stderr.print("[red]" + str(e) + "[/]")

        if "Permission denied" in str(e):
            stdout.print("\nHint: Is your user part of the docker group?")
        elif "invalid reference format" in str(e):
            stdout.print("\nHint: The input repository might not exist in the registry.")
        elif "400" in str(e):
            stdout.print("\nHint: The input parameters might not be valid.")
    except ContainerEngineError as e:
        stderr.print("[red]" + str(e) + "[/]")

# Call the main() when run as `python -m`
if __name__ == "__main__":
    main()



"""Repesents the Development Platform. The platform resources can be accessed through this interface.  
"""
from dem.core.core import Core
from dem.core.properties import __supported_dev_env_major_version__
from dem.core.exceptions import InvalidDevEnvJson
from dem.core.dev_env_catalog import DevEnvCatalogs
from dem.core.data_management import LocalDevEnvJSON, ConfigFile
from dem.core.container_engine import ContainerEngine
from dem.core.registry import Registries
from dem.core.tool_images import ToolImages
from dem.core.dev_env import DevEnv, DevEnv, DevEnv

class DevEnvSetup(Core):
    """ Representation of the Development Platform:
        - The available tool images.
        - The available Development Environments.
        - External resources.

        Class variables:
            _tool_images -- the available tool images
            _container_engine -- the container engine
            _regisitries -- managing the registries
            _config_file -- contains the DEM configuration
            update_tool_images_on_instantiation -- can be used to disable tool update if not needed
    """
    _tool_images = None
    _container_engine = None
    _registries = None
    _config_file = None
    update_tool_images_on_instantiation = True

    def _dev_env_json_version_check(self) -> None:
        """ Check that the json file is supported.

            The version is stored as string in the X.Y format.
            Raises an InvalidDevEnvJson exception, if the version is not supported.
        """
        dev_env_json_major_version = int(self.version.split('.', 1)[0])
        if dev_env_json_major_version != __supported_dev_env_major_version__:
            raise InvalidDevEnvJson("The dev_env.json version v1.0 is not supported.")

    def __init__(self, dev_env_json_deserialized: dict) -> None:
        """ Init the class, by creating a list of the Development Environments.

            Args:
                dev_env_json_deserialized -- a deserialized representation of the dev_env.json file 
        """
        self.version = dev_env_json_deserialized["version"]
        self._dev_env_json_version_check()
        self.local_dev_envs: list[DevEnv] = []
        self._dev_env_catalogs: DevEnvCatalogs | None = None

    @classmethod
    @property
    def tool_images(cls) -> ToolImages:
        """ The tool images.

            The class variable can be accessed through a getter, so the ToolImages() gets instantiated 
            only at the first access.

            Args:
                cls - the class object
        """
        if cls._tool_images is None:
            cls._tool_images = ToolImages(cls.container_engine, cls.registries,
                                          cls.update_tool_images_on_instantiation)
        return cls._tool_images

    @classmethod
    @property
    def container_engine(cls) -> ContainerEngine:
        """ The container engine.

            The class variable can be accessed through a getter, so the ContainerEngine() gets 
            instantiated only at the first access.

            Args:
                cls - the class object
        """
        if cls._container_engine is None:
            cls._container_engine = ContainerEngine()

        return cls._container_engine

    @classmethod
    @property
    def registries(cls) -> Registries:
        """ The registries.

            The class variable can be accessed through a getter, so the Registries() gets instantiated 
            only at the first access.

            Args:
                cls - the class object
        """
        if cls._registries is None:
            cls._registries = Registries(cls.container_engine, cls.config_file)

        return cls._registries

    @classmethod
    @property
    def config_file(cls) -> ConfigFile:
        """ The config file.

            The class variable can be accessed through a getter, so the ConfigFile() gets instantiated 
            only at the first access.

            Args:
                cls - the class object
        """
        if cls._config_file is None:
            cls._config_file = ConfigFile()

        return cls._config_file

    @property
    def dev_env_catalogs(self) -> DevEnvCatalogs:
        """ The Development Environment Catalogs.

            The DevEnvCatalogs() gets instantiated only at the first access.
        """
        if self._dev_env_catalogs is None:
            self._dev_env_catalogs = DevEnvCatalogs(self.config_file)

        return self._dev_env_catalogs

    def get_deserialized(self) -> dict:
        """ Create the deserialized json. 
        
            Return with the dev_env.json as a dict.
        """
        dev_env_json_deserialized = {}
        dev_env_json_deserialized["version"] = self.version
        dev_env_descriptors = []
        for dev_env in self.local_dev_envs:
            dev_env_descriptor = {}
            dev_env_descriptor["name"] = dev_env.name
            dev_env_descriptor["tools"] = dev_env.tools
            dev_env_descriptors.append(dev_env_descriptor)
        dev_env_json_deserialized["development_environments"] = dev_env_descriptors
        return dev_env_json_deserialized

    def get_dev_env_by_name(self, dev_env_name: str) -> ("DevEnv | DevEnv | None"):
        """ Get the Development Environment by name.
        
            Args:
                dev_env_name -- name of the Development Environment to get

            Return with the instance representing the Development Environment. If the Development 
            Environment doesn't exist in the setup, return with None.
        """
        for dev_env in self.local_dev_envs:
            if dev_env.name == dev_env_name:
                return dev_env

    def get_local_dev_env(self, catalog_dev_env: DevEnv) -> DevEnv | None:
        """ Get the local copy of the catalog's Dev Env if exists.

            Args:
                catalog_dev_env -- try to get the local copy of this catalog

            Return the local Dev Env object if available, None if not yet installed.
        """
        for local_dev_env in self.local_dev_envs:
            if catalog_dev_env.name == local_dev_env.name:
                return local_dev_env

class DevEnvLocalSetup(DevEnvSetup):
    def __init__(self) -> None:
        """ Store the local Development Environments.

        Extends the DevEnvSetup super class by populating the list of Development Environments with 
        DevEnv objects.
        """
        self.json = LocalDevEnvJSON()
        super().__init__(self.json.deserialized)

        for dev_env_descriptor in self.json.deserialized["development_environments"]:
            self.local_dev_envs.append(DevEnv(descriptor=dev_env_descriptor))

    def flush_to_file(self) -> None:
        """ Writes the deserialized json to the dev_env.json file."""
        # Get the up-to-date deserialized data.
        self.json.deserialized = self.get_deserialized()
        self.json.flush()

    def pull_images(self, tools: list) -> None:
        """ Pull images that are only present in the registry.
        
        Args:
            tools -- the tool images to pull (with any status, this function will filter the 
                     registry only ones)
        """

        # Remove the duplications and the locally already available tools
        tool_images_to_pull = set()
        for tool in tools:
            tool_image = tool["image_name" ] + ':' + tool["image_version"]
            if (tool["image_status"] == ToolImages.REGISTRY_ONLY):
                tool_images_to_pull.add(tool_image)

        for tool_image in tool_images_to_pull:
            self.user_output.msg("\nPulling image "+ tool_image, is_title=True)
            self.container_engine.pull(tool_image)


"""export CLI command implementation."""
# dem/cli/command/export_cmd.py


import re
from dem.core.platform import DevEnvLocalSetup
from dem.cli.console import stderr
import json, os

def check_is_directory(param: str):
     if None != param:
        return(os.path.isdir(param))     
     else:
        return False
     
def check_is_path_contains_spec_char(param: str):
    special_chars=re.compile('[~/\^]')
    if None != param:
        if None != special_chars.search(param):
            return True
        else:
            return False
    else:
        return False
    
def create_exported_dev_env_json(dev_env_name: str,dev_env_json: str,given_path: str):
    
    file_name=None
    file_path=None

    if True == check_is_directory(given_path):
        file_name=dev_env_name
        file_path=given_path+"/"
    elif True == check_is_path_contains_spec_char(given_path):        
        file_name=""
        file_path=given_path         
    elif None != given_path:
        file_name=given_path        
        file_path=""    
    else:
        file_name=dev_env_name        
        file_path=""
    
    exported_file = open(file_path+file_name, "w")        
    json.dump(dev_env_json, exported_file, indent=4)
    exported_file.close()        

def execute(dev_env_name: str, path_to_export: str) -> None:
    platform = DevEnvLocalSetup()
    dev_env_to_export = platform.get_dev_env_by_name(dev_env_name)

    if dev_env_to_export is not None:                
        create_exported_dev_env_json(dev_env_name,dev_env_to_export.__dict__,path_to_export)                
    else:
        stderr.print("[red]Error: The input Development Environment does not exist.[/]")


"""The pull command"""
# dem/cli/command/pull_cmd.py

from dem.core.tool_images import ToolImages
from dem.core.platform import DevEnvLocalSetup
from dem.core.dev_env import DevEnv, DevEnv
from dem.cli.console import stdout, stderr

def install_to_dev_env_json(local_dev_env: DevEnv | None, catalog_dev_env: DevEnv, 
                            platform: DevEnvLocalSetup) -> DevEnv:
    """Install the Dev Env descriptor to the dev_env.json.
    
    If the dev_env_local is None, the Dev Env is not yet installed locally. Install it by appending 
    the Dev Env descriptor to the dev_env.json file.
    If the Dev Env is already installed, but the descriptor is different, then overwrite the 
    dev_env.json element with the new one.
    If the Dev Env is already installed and both the descriptors are the same, then there is nothing
    to do. 
    Args:
        dev_env_local -- local Dev Env instance (none if not yet installed)
        dev_env_org -- organization's Dev Env instance
        dev_env_local_setup -- the local Dev Env setup
    Return with the dev_env_local instance.
    """
    if local_dev_env is None:
        # If not available, install it.
        local_dev_env = DevEnv(dev_env_to_copy=catalog_dev_env)
        platform.local_dev_envs.append(local_dev_env)
        platform.flush_to_file()
    elif local_dev_env.tools != catalog_dev_env.tools:
        # If already installed, but different, then overwrite it.
        local_dev_env.tools = catalog_dev_env.tools
        platform.flush_to_file()

    return local_dev_env

def execute(dev_env_name: str) -> None:
    # Get the organization's Dev Env if available.
    platform = DevEnvLocalSetup()
    catalog_dev_env: DevEnv | None = None

    if not platform.dev_env_catalogs.catalogs:
        stderr.print("[red]Error: No Development Environment Catalogs are available to pull the image from![/]")
        return

    for catalog in platform.dev_env_catalogs.catalogs:
        catalog_dev_env = catalog.get_dev_env_by_name(dev_env_name)
        if catalog_dev_env is not None:
            break
    else:
        stderr.print("[red]Error: The input Development Environment is not available for the organization.[/]")
        return

    local_dev_env = install_to_dev_env_json(platform.get_local_dev_env(catalog_dev_env), 
                                            catalog_dev_env, platform)

    # The local Dev Env setup contains the DevEnvOrg to install. Check the images' statuses
    local_dev_env.check_image_availability(platform.tool_images)
    platform.pull_images(local_dev_env.tools)
    # Check image availability again.
    image_statuses = local_dev_env.check_image_availability(platform.tool_images, 
                                                            update_tool_images=True)

    if image_statuses.count(ToolImages.LOCAL_AND_REGISTRY) == len(image_statuses):
        stdout.print("The [yellow]" + local_dev_env.name + "[/] Development Environment is ready!")
    else:
        stderr.print("The installation failed.")